#include "nrf24l01.h"
#include <string.h>

uint8_t child_pipe[6] = { RF24_RX_ADDR_P0, RF24_RX_ADDR_P1, RF24_RX_ADDR_P2, RF24_RX_ADDR_P3, RF24_RX_ADDR_P4, RF24_RX_ADDR_P5 };
uint8_t child_pipe_enable[6] = { RF24_ERX_P0, RF24_ERX_P1, RF24_ERX_P2, RF24_ERX_P3, RF24_ERX_P4, RF24_ERX_P5 };
uint8_t child_payload_size[6] = { RF24_RX_PW_P0, RF24_RX_PW_P1, RF24_RX_PW_P2, RF24_RX_PW_P3, RF24_RX_PW_P4, RF24_RX_PW_P5 };

extern char read_buf[512];
int debug_reg;

// initialize the module
int rf24_init() {
  
  rf24_payload_size = 32;
  rf24_ack_payload_available = 0;
  rf24_wide_band = 1;
  rf24_dynamic_payloads_enabled = 0;
  memset(rf24_pipe0_reading_address, 0, 5);
  
  // initialize pins here
  // set direction for CS (output) on RPB4 (pin 11)
  //TRISBbits.TRISB4 = 0;
  
  // direction for SCLK2 (output)
  //TRISBbits.TRISB15 = 0;
  
  // direction for MOSI2 (output)
  //TRISBbits.TRISB5 = 0;
  //RPB5R = 4;
  
  // direction for MISO2 (input)
  //TRISBbits.TRISB13 = 1;
  //SDI2R = 3;
  
  // set direction for CS (output) on RPB1 (pin 5)
  //TRISBbits.TRISB1 = 0;
  
  // set direction for CE (output) on RPB2 (pin 6)
  //TRISBbits.TRISB2 = 0;
  
    // set direction for CS (output) on RPB1 (pin 5)
  TRISBbits.TRISB1 = 0;
  
  // set direction for CE (output) on RPB2 (pin 6)
  TRISBbits.TRISB2 = 0;
  
  // direction for SCLK1 (output) on RPB14 (pin 25)
  TRISBbits.TRISB14 = 0;
  
  // direction for MOSI1 (output) on RPA1 (pin 3)
  TRISAbits.TRISA1 = 0;
  RPA1R = 3;
  
  // direction for MISO1 (input) on RPB8 (pin 17)
  TRISBbits.TRISB8 = 1;
  SDI1R = 4;
  
  // open SPI1
  // clock divisor of 160 required for 250khz operation. however clock divider
  // of 64 seems more stable.
  SpiChnClose(1);
  SpiChnOpen(1, SPI_OPEN_MSTEN | SPI_OPEN_MODE8 | SPICON_CKE | SPICON_ON | SPI_OPEN_ON, 16); //PBCLK/SPI_FREQ );
  
  rf24_delay_ms(100);
  
  _rf24_ce_low();
  _rf24_cs_high();
  
  // allow radio time to settle.
  rf24_delay_ms(5);
  
  rf24_write_single_register(RF24_SETUP_RETR, (4 << RF24_ARD) | (15 << RF24_ARC));
  
  rf24_set_amplifier_level(RF24_PA_MAX);
  int ret = rf24_set_data_rate(RF24_1MBPS);
  rf24_set_crc_length(RF24_CRC_16);
  rf24_write_single_register(RF24_DYNPD,0);
  rf24_write_single_register(RF24_NRF_STATUS, _BV(RF24_RX_DR) | _BV(RF24_TX_DS) | _BV(RF24_MAX_RT));
  rf24_set_channel(76);
  
  rf24_flush_rx();
  rf24_flush_tx();
  
  return ret;
}

// set the operating channel of the module
void rf24_set_channel(uint8_t channel) {
  uint8_t max_channel = 127;
  rf24_write_single_register(RF24_RF_CH, rf24_min(channel, max_channel));
}

// find out the operating channel of the module
uint8_t rf24_get_channel() {
  return rf24_read_single_register(RF24_RF_CH);
}

// sets the module data rate, returns 1 if successful, 0 if not
int rf24_set_data_rate(rf24_datarate_e speed) {
  
  uint8_t setup = rf24_read_single_register(RF24_RF_SETUP);
  
  // high and low 00 == 1mbs (default))
  setup &= ~(_BV(RF24_RF_DR_LOW) | _BV(RF24_RF_DR_HIGH));
  
  if (speed == RF24_250KBPS) {
    rf24_wide_band = 0;
    setup |= _BV(RF24_RF_DR_LOW);
  } else if (speed == RF24_2MBPS) {
    rf24_wide_band = 1;
    setup |= _BV(RF24_RF_DR_HIGH);
  } else {
    rf24_wide_band = 0;
  }
  
  rf24_write_single_register(RF24_RF_SETUP, setup);

  return (rf24_read_single_register(RF24_RF_SETUP) == setup ? 1 : 0);
}

// finds out the module data rate
rf24_datarate_e rf24_get_data_rate() {
  rf24_datarate_e result;
  uint8_t dr = rf24_read_single_register(RF24_RF_SETUP) & (_BV(RF24_RF_DR_LOW) | _BV(RF24_RF_DR_HIGH));
  
  if (dr == _BV(RF24_RF_DR_LOW)) {
    result = RF24_250KBPS;
  } else if (dr == _BV(RF24_RF_DR_HIGH)) {
    result = RF24_2MBPS;
  } else {
    result = RF24_1MBPS;
  }
  return result;
}

// sets the crc length of the packet
void rf24_set_crc_length(rf24_crclength_e length) {
  
  uint8_t config = rf24_read_single_register(RF24_CONFIG) & ~(_BV(RF24_CRCO) | _BV(RF24_EN_CRC));
  
  if (length == RF24_CRC_DISABLED) {}
  else if (length == RF24_CRC_8) {
    config |= _BV(RF24_EN_CRC);
  } else {
    config |= _BV(RF24_EN_CRC);
    config |= _BV(RF24_CRCO);
  }
  
  rf24_write_single_register(RF24_CONFIG, config);
}

rf24_crclength_e rf24_get_crc_length() {
  rf24_crclength_e result = RF24_CRC_DISABLED;
  uint8_t config = rf24_read_single_register(RF24_CONFIG) & ( _BV(RF24_CRCO) | _BV(RF24_EN_CRC)) ;

  if (config & _BV(RF24_EN_CRC)) {
    if (config & _BV(RF24_CRCO)) {
      result = RF24_CRC_16;
    } else {
      result = RF24_CRC_8;
    }
  }

  return result;
}

void rf24_disable_crc() {
  uint8_t disable = rf24_read_single_register(RF24_CONFIG) & ~_BV(RF24_EN_CRC) ;
  rf24_write_single_register(RF24_CONFIG, disable);
}

// change the amplifier power settings
void rf24_set_amplifier_level(rf24_pa_dbm_e level) {
  
  uint8_t setup = rf24_read_single_register(RF24_RF_SETUP);
  setup &= ~(_BV(RF24_RF_PWR_LOW) | _BV(RF24_RF_PWR_HIGH));
  
  if (level == RF24_PA_MAX) {
    setup |= (_BV(RF24_RF_PWR_LOW) | _BV(RF24_RF_PWR_HIGH));
  } else if (level == RF24_PA_HIGH) {
    setup |= (_BV(RF24_RF_PWR_HIGH));
  } else if (level == RF24_PA_LOW) {
    setup |= (_BV(RF24_RF_PWR_LOW));
  } else if (level == RF24_PA_MIN) {
    // do nothing
  } else {
    setup |= (_BV(RF24_RF_PWR_LOW) | _BV(RF24_RF_PWR_HIGH));
  }
  
  rf24_write_single_register(RF24_RF_SETUP, setup);
}

rf24_pa_dbm_e rf24_get_amplifier_level() {
  rf24_pa_dbm_e result = RF24_PA_ERROR;
  uint8_t power = rf24_read_single_register(RF24_RF_SETUP) & (_BV(RF24_RF_PWR_LOW) | _BV(RF24_RF_PWR_HIGH)) ;

  // switch uses RAM (evil!)
  if ( power == (_BV(RF24_RF_PWR_LOW) | _BV(RF24_RF_PWR_HIGH)) ) {
    result = RF24_PA_MAX ;
  } else if ( power == _BV(RF24_RF_PWR_HIGH) ) {
    result = RF24_PA_HIGH ;
  } else if ( power == _BV(RF24_RF_PWR_LOW) ) {
    result = RF24_PA_LOW ;
  } else {
    result = RF24_PA_MIN ;
  }

  return result ;
}

void rf24_start_listening() {
  
  rf24_write_single_register(RF24_CONFIG, rf24_read_single_register(RF24_CONFIG) | _BV(RF24_PWR_UP) | _BV(RF24_PRIM_RX));
  rf24_write_single_register(RF24_NRF_STATUS, _BV(RF24_RX_DR) | _BV(RF24_TX_DS) | _BV(RF24_MAX_RT));
  rf24_write_register(RF24_RX_ADDR_P0, rf24_pipe0_reading_address, 5);
  rf24_flush_rx();
  rf24_flush_tx();
  
  _rf24_ce_high();
  rf24_delay_ms(1); // actually only 130us needed
}

void rf24_stop_listening() {
  _rf24_ce_low();
  rf24_flush_rx();
  rf24_flush_tx();
}

void rf24_power_up() {
  rf24_write_single_register(RF24_CONFIG, rf24_read_single_register(RF24_CONFIG) | _BV(RF24_PWR_UP));
  rf24_delay_ms(5);
}

void rf24_power_down() {
  rf24_write_single_register(RF24_CONFIG, rf24_read_single_register(RF24_CONFIG) & ~_BV(RF24_PWR_UP));
}

void rf24_toggle_features() {
  _rf24_cs_low();
  rf24_spi_transfer(RF24_ACTIVATE);
  rf24_spi_transfer(0x73);
  _rf24_cs_high();
}

void rf24_set_retries(uint8_t delay, uint8_t count) {
  rf24_write_single_register(RF24_SETUP_RETR, ((delay & 0x0f) << RF24_ARD) | ((count & 0x0f) << RF24_ARC));
}

int rf24_available(uint8_t* pipe_num) {
  
  uint8_t status = rf24_get_status();
  uint8_t result = (status & _BV(RF24_RX_DR));
  
  if (result != 0) {
    if (pipe_num != NULL) {
      *pipe_num = (status >> RF24_RX_P_NO) & 7;
    }
    
    rf24_write_single_register(RF24_NRF_STATUS, _BV(RF24_RX_DR));
    
    // handle ack payload
    if (status & _BV(RF24_TX_DS)) {
      rf24_write_single_register(RF24_NRF_STATUS, _BV(RF24_TX_DS));
    }
  }
  
  return result;
}

int rf24_read(uint8_t* buf, uint8_t len) {
  rf24_read_payload(buf, len);
  return (rf24_read_single_register(RF24_FIFO_STATUS) & _BV(RF24_RX_EMPTY)) != 0 ? 1 : 0;
}

// writes a buffer to the RF24 module for transmission
int rf24_write(const uint8_t* buf, uint8_t len) {
  
  uint8_t result = 0;
  rf24_start_write(buf, len);
  
  uint8_t observe_tx;
  uint8_t status;
  
  do {
    status = rf24_read_register(RF24_OBSERVE_TX, &observe_tx, 1);
  } while (!(status & (_BV(RF24_TX_DS) | _BV(RF24_MAX_RT))));
  
  uint8_t tx_ok, tx_fail;
  rf24_what_happened(&tx_ok, &tx_fail, &rf24_ack_payload_available);
  
  result = tx_ok;
  if (rf24_ack_payload_available) {
    rf24_ack_payload_length = rf24_get_dynamic_payload_size();
  }
  
  rf24_power_down();
  rf24_flush_tx();
  
  return result;
}

int rf24_start_write(const uint8_t* buf, uint8_t len) {
  rf24_write_single_register(RF24_CONFIG, (rf24_read_single_register(RF24_CONFIG) | _BV(RF24_PWR_UP)) & ~_BV(RF24_PRIM_RX));
  rf24_delay_us(150);
  
  rf24_write_payload(buf, len);
  
  _rf24_ce_high();
  rf24_delay_us(15);
  _rf24_ce_low();
}

void rf24_open_write_pipe(const uint8_t* address) {
  rf24_write_register(RF24_RX_ADDR_P0, address, 5);
  rf24_write_register(RF24_TX_ADDR, address, 5);
  
  uint8_t max_payload_size = 32;
  if (rf24_payload_size > max_payload_size) rf24_payload_size = max_payload_size;
  
  rf24_write_single_register(RF24_RX_PW_P0, rf24_payload_size);
}

void rf24_open_read_pipe(uint8_t child, const uint8_t* address) {
  if (child == 0) {
    memcpy(rf24_pipe0_reading_address, address, 5);
  }
  
  if (child <= 6) {
    if (child < 2) {
      rf24_write_register(child_pipe[child], address, 5);
    } else {
      rf24_write_register(child_pipe[child], address, 1);
    }
    rf24_write_single_register(child_payload_size[child], rf24_payload_size);
    
    rf24_write_single_register(RF24_EN_RXADDR, rf24_read_single_register(RF24_EN_RXADDR) | _BV(child_pipe_enable[child]));
  }
}

// reads a single rf24 register into a buffer. does not set/unset CS
uint8_t rf24_read_register(uint8_t reg, uint8_t* buf, uint8_t len) {  
  uint8_t status;
  
  _rf24_cs_low();
  status = rf24_spi_transfer(RF24_R_REGISTER | (RF24_REGISTER_MASK & reg));
  while(len--) {
    *buf++ = rf24_spi_transfer(0xff);
  }
  _rf24_cs_high();
  
  return status;
}

// reads a single rf24 register. does not set/unset CS
uint8_t rf24_read_single_register(uint8_t reg) {
  uint8_t result;
  uint8_t status;

  _rf24_cs_low();
  status = rf24_spi_transfer(RF24_R_REGISTER | (RF24_REGISTER_MASK & reg));
  result = rf24_spi_transfer(0xff);
  _rf24_cs_high();
  
  return result;
}

// writes a buffer for a given rf24 register. does not set/unset CS
uint8_t rf24_write_register(uint8_t reg, const uint8_t* buf, uint8_t len) {
  uint8_t status;
  
  _rf24_cs_low();
  status = rf24_spi_transfer(RF24_W_REGISTER | (RF24_REGISTER_MASK & reg));
  while(len--) {
    rf24_spi_transfer(*buf++);
  }
  _rf24_cs_high();
  
  return status;
}

// writes a single register on the rf24. does not set/unset CS
uint8_t rf24_write_single_register(uint8_t reg, uint8_t value) {
  
  uint8_t status;
  uint8_t dummy;
  
  _rf24_cs_low();
  status = rf24_spi_transfer(RF24_W_REGISTER | (RF24_REGISTER_MASK & reg));
  dummy = rf24_spi_transfer(value);
  _rf24_cs_high();
  
  return status;
}

// writes payload into rf24. does not set/unset CS
uint8_t rf24_write_payload(const uint8_t* buf, uint8_t data_len) {
  
  uint8_t status;
  
  data_len = rf24_min(data_len, rf24_payload_size);
  uint8_t blank_len = rf24_dynamic_payloads_enabled ? 0 : rf24_payload_size - data_len;
  
  _rf24_cs_low();
  status = rf24_spi_transfer(RF24_W_TX_PAYLOAD);
    
  while(data_len--) {
    rf24_spi_transfer(*buf++);
  }
  
  while(blank_len--) {
    rf24_spi_transfer(0);
  }
  
  _rf24_cs_high();
  return status;
}

// reads payload from rf24. does not set/unset CS
uint8_t rf24_read_payload(uint8_t* buf, uint8_t data_len) {
  
  uint8_t status;
  if (data_len > rf24_payload_size) data_len = rf24_payload_size;
  uint8_t blank_len = rf24_dynamic_payloads_enabled ? 0 : rf24_payload_size - data_len;
  
  _rf24_cs_low();
  status = rf24_spi_transfer(RF24_R_RX_PAYLOAD);
  while(data_len--) {
    *buf++ = rf24_spi_transfer(0xff);
  }
  while(blank_len--) {
    rf24_spi_transfer(0xff);
  }
  _rf24_cs_high();
  return status;
}

// flush the transmit buffer.
uint8_t rf24_flush_tx() {
  
  uint8_t ret;
  _rf24_cs_low();
  ret = rf24_spi_transfer(RF24_FLUSH_TX);
  _rf24_cs_high();
  
  return ret;
}

// flush the receive buffer. note: does not set/unset CS
uint8_t rf24_flush_rx() {
  
  uint8_t ret;
  
  _rf24_cs_low();
  ret = rf24_spi_transfer(RF24_FLUSH_RX); 
  _rf24_cs_high();
  
  return ret;
}

// returns the status
uint8_t rf24_get_status() {
  
  uint8_t status;
  
  _rf24_cs_low();
  status = rf24_spi_transfer(RF24_NOP);
  _rf24_cs_high();
  
  return status;
}

void rf24_what_happened(uint8_t* tx_ok, uint8_t* tx_fail, uint8_t* rx_ready) {
  uint8_t status = rf24_write_single_register(RF24_NRF_STATUS, _BV(RF24_RX_DR) | _BV(RF24_TX_DS) | _BV(RF24_MAX_RT));
  *tx_ok = ((status & _BV(RF24_TX_DS)) != 0 ? 1 : 0);
  *tx_fail = ((status & _BV(RF24_MAX_RT)) != 0 ? 1 : 0);
  *rx_ready = ((status & _BV(RF24_RX_DR)) != 0 ? 1 : 0);
}

void rf24_enable_dynamic_payloads() {
  // Enable dynamic payload throughout the system
  rf24_write_single_register(RF24_FEATURE,rf24_read_single_register(RF24_FEATURE) | _BV(RF24_EN_DPL));

  // If it didn't work, the features are not enabled
  if (!rf24_read_single_register(RF24_FEATURE)) {
    // So enable them and try again
    rf24_toggle_features();
    rf24_write_single_register(RF24_FEATURE,rf24_read_single_register(RF24_FEATURE) | _BV(RF24_EN_DPL) );
  }

  // Enable dynamic payload on all pipes
  //
  // Not sure the use case of only having dynamic payload on certain
  // pipes, so the library does not support it.
  rf24_write_single_register(RF24_DYNPD,rf24_read_single_register(RF24_DYNPD)
    | _BV(RF24_DPL_P5) | _BV(RF24_DPL_P4) | _BV(RF24_DPL_P3) | _BV(RF24_DPL_P2) | _BV(RF24_DPL_P1) | _BV(RF24_DPL_P0));

  rf24_dynamic_payloads_enabled = 1;
}

uint8_t rf24_get_dynamic_payload_size() {
  uint8_t result;
  
  _rf24_cs_low();
  rf24_spi_transfer(RF24_R_RX_PL_WID);
  result = rf24_spi_transfer(0xff);
  _rf24_cs_high();
  
  return result;
}

void rf24_enable_ack_payload() {
  rf24_write_single_register(RF24_FEATURE, rf24_read_single_register(RF24_FEATURE) | _BV(RF24_EN_ACK_PAY) | _BV(RF24_EN_DPL));

  // If it didn't work, the features are not enabled
  if (rf24_read_single_register(RF24_FEATURE) == 0) {
    // So enable them and try again
    rf24_toggle_features();
    rf24_write_single_register(RF24_FEATURE, rf24_read_single_register(RF24_FEATURE) | _BV(RF24_EN_ACK_PAY) | _BV(RF24_EN_DPL));
  }
  //
  // Enable dynamic payload on pipes 0 & 1
  //
  rf24_write_single_register(RF24_DYNPD, rf24_read_single_register(RF24_DYNPD) | _BV(RF24_DPL_P1) | _BV(RF24_DPL_P0));
}

int rf24_test_carrier() {
  return (rf24_read_single_register(RF24_CD) & 1);
}

int rf24_test_rpd() {
  return (rf24_read_single_register(RF24_RPD) & 1);
}

void rf24_set_auto_ack_all(int enable) {
  if (enable > 0) {
    rf24_write_single_register(RF24_EN_AA, 63); // b111111
  } else {
    rf24_write_single_register(RF24_EN_AA, 0);
  }
  
}

void rf24_set_auto_ack(uint8_t pipe, int enable) {
  if (pipe <= 6) {
    uint8_t en_aa = rf24_read_single_register(RF24_EN_AA) ;
    if(enable > 0) {
      en_aa |= _BV(pipe) ;
    } else {
      en_aa &= ~_BV(pipe) ;
    }
    rf24_write_single_register(RF24_EN_AA, en_aa );
  }
}

void rf24_write_ack_payload(uint8_t pipe, uint8_t* buf, uint8_t len) {

  _rf24_cs_low();
  rf24_spi_transfer(RF24_W_ACK_PAYLOAD | (pipe & 7));
  const uint8_t max_payload_size = 32;
  uint8_t data_len = min(len,max_payload_size);
  while (data_len--) {
    rf24_spi_transfer(*buf++);
  }
  _rf24_cs_high();
}

int rf24_is_ack_payload_available() {
  int result = rf24_ack_payload_available;
  rf24_ack_payload_available = 0;
  return result;
}

void rf24_print_address_register(char* dest, const char* name, uint8_t reg, uint8_t qty) {
  
  char temp[128];
  while(qty--) {
    uint8_t buffer[5];
    rf24_read_register(reg++, buffer, sizeof(buffer));
    
    sprintf(temp, "%s %02x %02x %02x %02x %02x", name, buffer[0], buffer[1], buffer[2], buffer[3], buffer[4]);
    strcat(dest, temp);
  }
}

void rf24_print_byte_register(char* dest, const char* name, uint8_t reg, uint8_t qty) {
  char temp[128];
  char temp2[32];
  sprintf(temp, "%s:", name);
  while(qty--) {
    sprintf(temp2, "%02x ", rf24_read_single_register(reg++));
    strcat(temp, temp2);
  }
  strcat(dest, temp);
}

void rf24_print_status(char* dest, uint8_t status) {
  char temp[128];
  sprintf(temp, "status:%02x RX_DR=%x TX_DR=%x MAX_RT=%x RX_P_NO=%x TX_FULL=%x",
    status,
    (status & _BV(RF24_RX_DR))?1:0,
    (status & _BV(RF24_TX_DS))?1:0,
    (status & _BV(RF24_MAX_RT))?1:0,
    ((status >> RF24_RX_P_NO) & 7),
    (status & _BV(RF24_TX_FULL))?1:0);
  strcat(dest, temp);
}

void rf24_print_details(char* dest) {
  
  rf24_print_status(dest, rf24_get_status());
  rf24_print_address_register(dest, " RX_ADDR_P0-1:", RF24_RX_ADDR_P0, 2);
  rf24_print_byte_register(dest, " RX_ADDR_P2-5:", RF24_RX_ADDR_P2, 4);
  rf24_print_address_register(dest, " TX_ADDR:", RF24_TX_ADDR, 1);
  rf24_print_byte_register(dest, " RX_PW_P0-6:", RF24_RX_PW_P0,6);
  rf24_print_byte_register(dest, " EN_AA:",RF24_EN_AA, 1);
  rf24_print_byte_register(dest, " EN_RXADDR:", RF24_EN_RXADDR, 1);
  rf24_print_byte_register(dest, " RF_CH:", RF24_RF_CH, 1);
  rf24_print_byte_register(dest, " RF_SETUP:", RF24_RF_SETUP, 1);
  rf24_print_byte_register(dest, " CONFIG:", RF24_CONFIG, 1);
  rf24_print_byte_register(dest, " DYNPD FEATURE:", RF24_DYNPD, 2);
  
  rf24_datarate_e datarate = rf24_get_data_rate();
  rf24_pa_dbm_e power_level = rf24_get_amplifier_level();
  rf24_crclength_e crc_len = rf24_get_crc_length();
  
  strcat(dest, " datarate: ");
  switch (datarate) {
  case RF24_1MBPS:
    strcat(dest, "1mbps ");
    break;
  case RF24_2MBPS:
    strcat(dest, "2mbps");
    break;
  default:
    strcat(dest, "250kbps");
    break;
  }
  
  strcat(dest, " power level: ");
  switch(power_level) {
    RF24_PA_MIN:
      strcat(dest, "min");
      break;
    
    RF24_PA_LOW:
      strcat(dest, "low");
      break;
    RF24_PA_HIGH:
      strcat(dest, "high");
      break;
    RF24_PA_MAX:
      strcat(dest, "max");
      break;
  default:
    strcat(dest, "error");
    break;
  }
  
  strcat(dest, " crc: ");
  switch(crc_len) {
  case RF24_CRC_DISABLED:
    strcat(dest, "disabled");
    break;
  case RF24_CRC_8:
    strcat(dest, "8bit");
    break;
  case RF24_CRC_16:
    strcat(dest, "16bit");
    break;
  default:
    strcat(dest, "error");
    break;
  }
  
}

uint8_t rf24_spi_transfer(uint8_t val) {
  
  uint8_t ret = 0;
  uint8_t ret2 = 0;
  while(TxBufFullSPI1());
  SPI1BUF = val;
  
  while(!SPI1STATbits.SPIRBF);
  ret = SPI1BUF;
  return ret;
}

void rf24_delay_ms(unsigned long i){
  unsigned int j;
  j = dTime_ms * i;
  WriteCoreTimer(0);
  while (ReadCoreTimer() < j);
}

void rf24_delay_us(unsigned long i) {
  unsigned int j;
  j = dTime_us * i;
  WriteCoreTimer(0);
  while (ReadCoreTimer() < j);
}
